import email
import random
import smtplib
import time

import loremipsum
from pvHelpers.mail.email import EmailHelpers


class GetSMTPConnetion(object):
    def __init__(self, config, login_as=(None, None)):
        self.config = config
        self.conn = None
        (self.username, self.password) = login_as

    def __enter__(self):
        self.conn = smtplib.SMTP(self.config.host, self.config.port, timeout=10)
        if self.config.use_tls:
            self.conn.starttls()

        if self.username and self.password:
            self.conn.login(self.config.username, self.config.password)
            self.conn.ehlo_or_helo_if_needed()
        elif self.config.username and self.config.password:
            self.conn.login(self.config.username, self.config.password)
            self.conn.ehlo_or_helo_if_needed()

        return self.conn

    def __exit__(self, type_, value, traceback):
        self.conn.close()


def test_smtp_connection(config):
    with GetSMTPConnetion(config) as conn:
        print conn.helo()
        print conn.verify(config.username)


def send_test_mail(config, message):
    with GetSMTPConnetion(config) as conn:
        # TODO assert return codes
        conn.verify(config.username)
        return conn.sendmail(config.username, [config.send_to_user], message)


def create_test_mime(sender_id, recipient_ids):
    n = time.time()
    return '''
        X-PreVeil-Msg
        Message-ID: <{}>
        From: {} <{}>
        To: {}
        Date: {}

        {}
        '''.format(
            EmailHelpers.newMessageId(),
            sender_id, sender_id,
            ', '.join(map(lambda i: '{} <{}>'.format(i, i), recipient_ids)),
            u'{}'.format(email.utils.formatdate(n - random.randint(100000, 999999999))),
            loremipsum.generate_paragraph()[2])



import os
import random
import time
import uuid

import loremipsum
from mail_relay.secure_channel.websocket_wrapper import (WSError, WSNoData,
                                                         WSWebSocketWrapper)
from pvHelpers.crypto import PVKeyFactory
from pvHelpers.mail.email import EmailHelpers
from pvHelpers.user import LocalDevice, LocalUser, OrganizationInfo
from pvHelpers.utils import jdumps, NOT_ASSIGNED, randUnicode
import requests


def rand_user_id():
    return u'{}{}@preveil.test'.format(randUnicode(), os.getpid())


def test_local_device():
    return LocalDevice(
        unicode(uuid.uuid4()),
        randUnicode(10),
        PVKeyFactory.newUserKey(0),
        {},
        u'active')


def test_local_user():
    return LocalUser(
        rand_user_id(),
        random.randint(0, 1000),
        randUnicode(10),
        unicode(uuid.uuid4()),
        OrganizationInfo(
            unicode(uuid.uuid4()), randUnicode(10), randUnicode(10), random.choice([u'standard', u'admin'])),
        NOT_ASSIGNED(),
        randUnicode(10),
        [PVKeyFactory.newUserKey(i-1) for i in xrange(random.randint(1, 10))],
        test_local_device())


def create_test_mime(sender_id=None, recipient_ids=None):
    recipient_ids = recipient_ids or [rand_user_id() for _ in xrange(random.randint(1, 5))]
    sender_id = sender_id or rand_user_id()
    return '''
        X-PreVeil-Msg
        Message-ID: <{}>
        From: {} <{}>
        To: {}

        {}
        '''.format(
            EmailHelpers.newMessageId(),
            sender_id, sender_id,
            ', '.join(map(lambda i: '{} <{}>'.format(i, i), recipient_ids)),
            loremipsum.generate_paragraph()[2])


def devices_equal(d1, d2):
    return all([
        d1.id.lower() == d2.id.lower(),
        d1.name.lower() == d2.name.lower(),
        d1.status == d2.status,
        d1.metadata == d2.metadata,
        d1.key == d2.key
    ])


def org_infos_equal(o1, o2):
    return all([
        o1.org_id.lower() == o2.org_id.lower(),
        o1.role.lower() == o2.role.lower(),
        o1.dept_name.lower() == o2.dept_name.lower(),
        o1.org_name.lower() == o2.org_name.lower()
    ])


def users_equal(u1, u2):
    return all([
        u1.user_id.lower() == u2.user_id.lower(),
        u1.mail_cid.lower() == u2.mail_cid.lower(),
        u1.display_name.lower() == u2.display_name.lower(),
        org_infos_equal(u1.org_info, u2.org_info),
        all([k1 == k2 for k1 in u1.user_keys for k2 in u2.user_keys if k1.key_version == k2.key_version]),
        devices_equal(u1.device, u2.device)
    ])


def wait_for_data(socket_wrapper, timeout=2):
    current_time = time.time()
    while (time.time() - current_time) < timeout:
        data = socket_wrapper.receive()
        assert not isinstance(data, WSError)
        if isinstance(data, WSNoData):
            time.sleep(0.1)
            continue
        # data is of type WSData
        return data.message

    assert False


def crypto_create_account(crypto_host, crypto_port, user_id):
    resp = requests.put(
        u'http://{}:{}/put/test/account/{}'.format(crypto_host, crypto_port, user_id),
        headers={'Content-Type': 'application/json', 'HOST': '127.0.0.1:2002'},
        data=jdumps({'name': randUnicode(5)})
    )
    resp.raise_for_status()
    return resp.json()


def crypto_send_user_key(crypto_host, crypto_port, for_user):
    conn = WSWebSocketWrapper('ws://{}:{}/get/account/{}/key/send'.format(crypto_host, crypto_port, for_user), 4, 0.1)

    conn.send(jdumps({'user_id': for_user, 'action': 'init'}))
    result = wait_for_data(conn)
    assert result['status'] == 'ok'

    def verify_pin(pin):
        conn.send(jdumps({'user_id': for_user, 'action': 'verify', 'pin': pin}))
        conn.close()

    return verify_pin



from pvHelpers.logger import g_log
from pvHelpers.utils import EncodingException, jloads, utf8Decode
import websocket


WS_READ_TIMEOUT = 0.1
WS_CONN_TIMEOUT = 4


class WSData:
    def __init__(self, data):
        self.message = data


class WSNoData:
    def __init__(self):
        pass


class WSError:
    def __init__(self, message=u''):
        self.message = message


class WSWebSocketWrapper:
    '''client-side low level websocket wrapper'''

    def __init__(self, path, conn_timeout=WS_CONN_TIMEOUT, read_timeout=WS_READ_TIMEOUT):
        self.websocket = websocket.create_connection(path, timeout=conn_timeout)
        self.websocket.settimeout(read_timeout)

    def receive(self):
        try:
            opcode, data = self.websocket.recv_data(control_frame=True)
        except websocket.WebSocketTimeoutException as e:
            return WSNoData()
        except websocket.WebSocketException as e:
            g_log.exception(e)
            return WSError(u'WSWebSocketWrapper.receive: excepion: {}'.format(e))

        if opcode in [websocket.ABNF.OPCODE_PING, websocket.ABNF.OPCODE_PONG]:
            return WSNoData()
        elif opcode == websocket.ABNF.OPCODE_CLOSE:
            return WSError(u'WSWebSocketWrapper.receive: Close frame recevied')
        elif opcode == websocket.ABNF.OPCODE_BINARY:
            return WSError('WSWebSocketWrapper.receive: Binary frame not expected')

        if data is None:
            return WSError(u'WSWebSocketWrapper.receive: Data can\'t be None')

        try:
            data = jloads(utf8Decode(data))
        except EncodingException as e:
            g_log.exception(e)
            return WSError(u'WSWebSocketWrapper.receive: Failed to parse json')

        return WSData(data)

    def send(self, data):
        self.websocket.send(data)

    def close(self):
        try:
            return self.websocket.close()
        except (ValueError, websocket.WebSocketException) as e:
            g_log.exception(e)



import os

from mail_relay.config import Config, CSConfig, IMAPConfig, SMTPConfig
import pytest


@pytest.fixture(scope='session')
def test_config():
    return Config(
        SMTPConfig(os.environ['SMTP_HOST'], int(os.environ['SMTP_PORT']), None, None, False, None),
        IMAPConfig(os.environ['IMAP_HOST'], int(os.environ['IMAP_PORT']), None, None, False),
        CSConfig(os.environ['CS_HOST'], int(os.environ['CS_PORT']), False))




class ExporterBucket:
    NAME = u'exporter'
    CURRENT_VERSION = 0

    @staticmethod
    def key():
        return 'exporter'

    @staticmethod
    def serialize(user_id, account_version, bucket_version):
        if bucket_version == 0:
            return ExporterBucket._serialize_v0(user_id, account_version)
        else:
            raise ValueError(u'unsupported bucket bucket_version {}'.format(bucket_version))

    @staticmethod
    def deserialize(blob, bucket_version):
        if bucket_version == 0:
            return ExporterBucket._deserialize_v0(blob)
        else:
            raise ValueError(u'unsupported bucket bucket_version {}'.format(bucket_version))

    @staticmethod
    def _serialize_v0(user_id, account_version):
        return u'{};{}'.format(account_version, user_id)

    @staticmethod
    def _deserialize_v0(blob):
        a = blob.split(';', 1)
        return a[0], a[1]



from pvHelpers.crypto import (ASYMM_KEY_PROTOCOL_VERSION,
                              PVKeyFactory,
                              SIGN_KEY_PROTOCOL_VERSION,
                              SYMM_KEY_PROTOCOL_VERSION)
from pvHelpers.crypto.box import AsymmBox, AsymmBoxV3
from pvHelpers.crypto.utils import HexEncode, Sha512Sum
from pvHelpers.utils import b64dec, b64enc, utf8Decode, utf8Encode


class DHProtocols(object):
    V1 = 1  # usage of ASYMM_KEY_PROTOCOL_VERSION.V2
    V2 = 2  # usage of ASYMM_KEY_PROTOCOL_VERSION.V3

    Current = V2


class KeyExchangeManager(object):
    __supported_protocols__ = [DHProtocols.V1, DHProtocols.V2]

    def __init__(self, protocol_version, ephemeral_key):
        self.ephemeral_key = ephemeral_key
        self.protocol_version = protocol_version
        self.other_ephemeral_key = None

    @classmethod
    def new(cls, protocol_version=None):
        if protocol_version is None:
            protocol_version = DHProtocols.Current

        if protocol_version == DHProtocols.V1:
            ephemeral_key = PVKeyFactory.newUserKey(
                0,
                encryption_key=PVKeyFactory.newAsymmKey(ASYMM_KEY_PROTOCOL_VERSION.V2),
                signing_key=PVKeyFactory.newSignKey(SIGN_KEY_PROTOCOL_VERSION.V1))
        elif protocol_version == DHProtocols.V2:
            ephemeral_key = PVKeyFactory.newUserKey(
                0,
                encryption_key=PVKeyFactory.newAsymmKey(ASYMM_KEY_PROTOCOL_VERSION.V3),
                signing_key=PVKeyFactory.newSignKey(SIGN_KEY_PROTOCOL_VERSION.V3))
        else:
            raise NotImplementedError('Unknown DH protocol {}'.format(protocol_version))

        return KeyExchangeManager(protocol_version, ephemeral_key)

    @staticmethod
    def ephemeral_key_hash(serialized_key):
        encoded = utf8Encode(serialized_key)
        digest = Sha512Sum(encoded)
        return HexEncode(digest)

    @property
    def pin(self):
        if self.other_ephemeral_key is None:
            raise RuntimeError('other side\'s public key is not set')

        if self.protocol_version == DHProtocols.V1:
            box = AsymmBox(self.ephemeral_key.encryption_key, self.other_ephemeral_key.public_key)
            return box.getPin()
        elif self.protocol_version == DHProtocols.V2:
            shared_key = AsymmBoxV3.get_shared_key(
                self.ephemeral_key.encryption_key, self.other_ephemeral_key.public_key)
            digest = Sha512Sum(shared_key)
            return HexEncode(digest)[:8]

        raise NotImplementedError('Unknown transfer protocol {}'.format(self.protocol_version))

    def encrypt(self, message):
        if self.other_ephemeral_key is None:
            raise RuntimeError('other side\'s public key is not set')

        if self.protocol_version == DHProtocols.V1:
            box = AsymmBox(self.ephemeral_key.encryption_key, self.other_ephemeral_key.public_key)
            return b64enc(box.encrypt(utf8Encode(message), is_text=True))
        elif self.protocol_version == DHProtocols.V2:
            shared_key = AsymmBoxV3.get_shared_key(
                self.ephemeral_key.encryption_key, self.other_ephemeral_key.public_key)
            key = PVKeyFactory.newSymmKey(SYMM_KEY_PROTOCOL_VERSION.V1, secret=shared_key)
            cipher = key.encrypt(utf8Encode(message))
            return b64enc(cipher)

        raise NotImplementedError('Unknown transfer protocol {}'.format(self.protocol_version))

    def decrypt(self, cipher):
        if self.other_ephemeral_key is None:
            raise RuntimeError('other side\'s public key is not set')

        if self.protocol_version == DHProtocols.V1:
            box = AsymmBox(self.ephemeral_key.encryption_key, self.other_ephemeral_key.public_key)
            return utf8Decode(box.decrypt(b64dec(cipher)))
        elif self.protocol_version == DHProtocols.V2:
            shared_key = AsymmBoxV3.get_shared_key(
                self.ephemeral_key.encryption_key, self.other_ephemeral_key.public_key)
            key = PVKeyFactory.newSymmKey(SYMM_KEY_PROTOCOL_VERSION.V1, secret=shared_key)
            plaintext = key.decrypt(b64dec(cipher))
            return utf8Decode(plaintext)

        raise NotImplementedError('Unknown transfer protocol {}'.format(self.protocol_version))



import time

from pvHelpers.utils import jdumps

from .key_exchange_manager import KeyExchangeManager
from .websocket_wrapper import WSData, WSError, WSWebSocketWrapper


class SecureMessageSendingStates():
    NOT_CONNECTED = 1
    WAITING_FOR_HANDSHAKE = 2
    WAITING_FOR_HASH = 3
    WAITING_FOR_PUBLIC_KEY = 4
    WAITING_FOR_VERIFICATION = 5
    WAITING_FOR_OK = 6
    WAITING_FOR_ENCRYPTED_MESSAGE = 7
    DONE = 8


class TransferTypes(object):
    TRANSFER = u'key_transfer'


class SecureMessageManager(object):
    __handshake_protocol__ = 1

    def __init__(self, relay_server_url, user_id, connect_to):
        self.relay_server_url = relay_server_url
        self.user_id = user_id
        self.connect_to = connect_to

        self.state = SecureMessageSendingStates.NOT_CONNECTED
        self.relay_websocket = None
        self.stopped = False

        self.other_ephemeral_public_key_hash = None

    def _advance_state(self):
        raise NotImplementedError()

    def connect_to_relay_server(self):
        if self.state != SecureMessageSendingStates.NOT_CONNECTED:
            raise RuntimeError('Unexpected state {}'.format(self.state))

        self.relay_websocket = WSWebSocketWrapper(u'{}/relay_ws'.format(self.relay_server_url))

    def send_handshake(self):
        user_id = self.user_id if self.__transfer_side__ == 'connect' or self.connect_to is None else self.connect_to
        return self.relay_websocket.send(jdumps({
            'handshake_protocol': self.__handshake_protocol__,
            'accept_protocols': KeyExchangeManager.__supported_protocols__,
            'user_id': user_id,
            'connect_to': self.connect_to,
            'action': '{}_handshake'.format(self.__transfer_side__),
            'type': self.transfer_type
        }))

    def init_dh(self, dh_manager):
        self._dh_manager = dh_manager
        user_id = self.user_id if self.__transfer_side__ == 'connect' or self.connect_to is None else self.connect_to
        self.relay_websocket.send(jdumps({
            'user_id': user_id,
            'action': self.__transfer_side__,
            'type': self.transfer_type,
            'public_key_hash': self._dh_manager.ephemeral_key_hash(
                self._dh_manager.ephemeral_key.public_user_key.serialize()),
            'protocol_version': self.protocol_version
        }))

    def resolve_common_protocol(self, handshake_message):
        common_protocols = \
            set(handshake_message['accept_protocols']) \
            .intersection(set(KeyExchangeManager.__supported_protocols__))

        if len(common_protocols) == 0:
            return None
        return max(common_protocols)

    def read_relay_socket(self):
        while True:
            data = self.relay_websocket.receive()
            if isinstance(data, WSData):
                if data.message['action'] != 'ok':
                    if data.message['data']['status'] == u'forbidden':
                        raise ValueError(u'key transfer forbidden by organization')
                return data.message['data']
            elif isinstance(data, WSError):
                raise ValueError(data.message)

            time.sleep(0.05)

    def stop(self):
        if self.stopped:
            return
        self.stopped = True

        if self.relay_websocket:
            self.relay_websocket.close()



import imaplib


class GetIMAPConnetion(object):
    def __init__(self, config):
        self.imap_config = config
        self.imap_conn = None

    def __enter__(self):
        self.imap_conn = imaplib.IMAP4(self.imap_config.host, self.imap_config.port)
        # do loging
        return self.imap_conn

    def __exit__(self, type_, value, traceback):
        self.imap_conn.close()



from mail_relay.config import Config
from pvHelpers.utils import jdumps, jloads


class ConfigBucket:
    NAME = u'config'
    CURRENT_VERSION = 0

    @staticmethod
    def key():
        return 'config'

    @staticmethod
    def serialize(config, bucket_version):
        if bucket_version == 0:
            return ConfigBucket._serialize_v0(config)
        else:
            raise ValueError(u'unsupported bucket bucket_version {}'.format(bucket_version))

    @staticmethod
    def deserialize(blob, bucket_version):
        if bucket_version == 0:
            return ConfigBucket._deserialize_v0(blob)
        else:
            raise ValueError(u'unsupported bucket bucket_version {}'.format(bucket_version))

    @staticmethod
    def _serialize_v0(config):
        return jdumps(config.to_json())

    @staticmethod
    def _deserialize_v0(blob):
        j = jloads(blob)
        return Config.from_json(j)



import random

from mail_relay.config import Config, CSConfig, IMAPConfig, SMTPConfig
from mail_relay.store.handlers import (read_config, read_key, read_keys,
                                       read_users, update_config, write_key,
                                       write_user)
from pvHelpers.crypto import PVKeyFactory
from pvHelpers.utils import randUnicode

from ..utils import rand_user_id, test_local_user, users_equal


def key_bucket_test(store):
    for _ in xrange(random.randint(0, 10)):
        test_user = rand_user_id()
        account_version = random.randint(0, 1000)
        key_version_count = random.randint(0, 40)
        for i in xrange(key_version_count):
            k = PVKeyFactory.newUserKey(i)
            write_key(test_user, account_version, k, store)
            k2 = read_key(test_user, account_version, i, store)
            assert k2 == k

        keys = read_keys(test_user, account_version, store)
        assert len(keys) == key_version_count


def user_bucket_test(store):
    test_users = []
    for _ in xrange(random.randint(0, 10)):
        test_users.append(test_local_user())
        write_user(test_users[-1], store)

    users = read_users(store)
    for tu in test_users:
        assert users_equal(tu, users[(tu.account_version, tu.user_id)])


def config_bucket_test(store):
    config = read_config(store)
    assert config is None

    test_config = Config(
        SMTPConfig(randUnicode(10), random.randint(0, 55555), randUnicode(10), randUnicode(10), rand_user_id(), False),
        IMAPConfig(randUnicode(10), random.randint(0, 55555), randUnicode(10), randUnicode(10), False),
        CSConfig(randUnicode(10), random.randint(0, 55555), False)
    )
    update_config(test_config, store)
    config = read_config(store)
    assert config.smtp.password == test_config.smtp.password

    test_config.smtp.password = randUnicode(10)
    update_config(test_config, store)
    config2 = read_config(store)
    assert config2.smtp.password == test_config.smtp.password



from pvHelpers.store import GetDBSession
from pvHelpers.utils import CaseInsensitiveDict, MergeDicts

from .config import ConfigBucket
from .exporter import ExporterBucket
from .key import KeyBucket
from .user import UserBucket


def write_key(user_id, account_version, key, store_path):
    with GetDBSession(store_path) as session:
        try:
            session.begin()
            res = session.execute(
                '''
                INSERT INTO Store (bucket, bucket_version, key, value)
                VALUES (:bucket, :version, :key, :value)
                ''', {
                    'bucket': KeyBucket.NAME,
                    'version': KeyBucket.CURRENT_VERSION,
                    'key': KeyBucket.key(user_id, account_version, key.key_version),
                    'value': KeyBucket.serialize(key, KeyBucket.CURRENT_VERSION)
                }
            )
            session.commit()
            return res.lastrowid
        except Exception:
            session.rollback()
            raise


def read_key(user_id, account_version, key_version, store_path):
    with GetDBSession(store_path) as session:
        result = session.execute(
            '''
            SELECT id, value, bucket_version FROM Store
            WHERE bucket = :bucket AND key = :key
            ''', {
                'bucket': KeyBucket.NAME,
                'key': KeyBucket.key(user_id, account_version, key_version)
            }
        )
        (_, blob, b_v) = result.fetchall()[0]

    return KeyBucket.deserialize(blob, b_v)


def read_keys(user_id, account_version, store_path):
    with GetDBSession(store_path) as session:
        result = session.execute(
            '''
            SELECT id, value, bucket_version FROM Store
            WHERE bucket = :bucket AND key LIKE :key
            ''', {
                'bucket': KeyBucket.NAME,
                'key': KeyBucket.key(user_id, account_version, '%')
            }
        )
        results = result.fetchall()

    return [KeyBucket.deserialize(blob, b_v) for (_, blob, b_v) in results]


def write_user(user, store_path):
    with GetDBSession(store_path) as session:
        try:
            session.begin()
            res = session.execute(
                '''
                INSERT INTO Store (bucket, bucket_version, key, value)
                VALUES (:bucket, :version, :key, :value)
                ''', {
                    'bucket': UserBucket.NAME,
                    'version': UserBucket.CURRENT_VERSION,
                    'key': UserBucket.key(user.user_id, user.account_version),
                    'value': UserBucket.serialize(user, UserBucket.CURRENT_VERSION)
                }
            )
            session.commit()
            return res.lastrowid
        except Exception:
            session.rollback()
            raise


def delete_user(user_id, store_path):
    with GetDBSession(store_path) as session:
        try:
            session.begin()
            session.execute(
                '''
                DELETE FROM Store
                WHERE bucket=:bucket
                AND key LIKE :key
                ''', {
                    'bucket': UserBucket.NAME,
                    'key': UserBucket.key(user_id, '%')
                }
            )
            session.execute(
                '''
                DELETE FROM Store
                WHERE bucket=:bucket
                AND key LIKE :key
                ''', {
                    'bucket': KeyBucket.NAME,
                    'key': KeyBucket.key(user_id, '%', '%')
                }
            )
            session.commit()
        except Exception:
            session.rollback()
            raise


# TODO: add single fetch api
def read_user(user_id, account_version, store_path):
    pass


def read_users(store_path):
    with GetDBSession(store_path) as session:
        result = session.execute(
            '''
            SELECT id, value, bucket_version FROM Store
            WHERE bucket = :bucket
            ''', {
                'bucket': UserBucket.NAME
            }
        )
        results = result.fetchall()

    users = [UserBucket.deserialize(blob, b_v) for (_, blob, b_v) in results]

    for u in users:
        u.user_keys = read_keys(u.user_id, u.account_version, store_path)

    latest_versions = CaseInsensitiveDict()
    for u in users:
        if (-1, u.user_id) in latest_versions and \
           latest_versions[(-1, u.user_id)].account_version < u.account_version:

            latest_versions[(-1, u.user_id)] = u

        else:
            latest_versions[(-1, u.user_id)] = u

    return CaseInsensitiveDict(MergeDicts(
        {(u.account_version, u.user_id): u for u in users},
        latest_versions
    ))


def update_config(config, store_path):
    with GetDBSession(store_path) as session:
        try:
            session.begin()
            res = session.execute(
                '''
                INSERT OR REPLACE INTO Store (bucket, bucket_version, key, value)
                VALUES (:bucket, :version, :key, :value)
                ''', {
                    'bucket': ConfigBucket.NAME,
                    'version': ConfigBucket.CURRENT_VERSION,
                    'key': ConfigBucket.key(),
                    'value': ConfigBucket.serialize(config, UserBucket.CURRENT_VERSION)
                }
            )
            session.commit()
            return res.lastrowid
        except Exception:
            session.rollback()
            raise


def read_config(store_path):
    with GetDBSession(store_path) as session:
        result = session.execute(
            '''
            SELECT id, value, bucket_version FROM Store
            WHERE bucket = :bucket AND key = :key
            ''', {
                'bucket': ConfigBucket.NAME,
                'key': ConfigBucket.key()
            }
        )
        results = result.fetchall()

    if len(results) == 1:
        (_, blob, b_v) = results[0]
        return ConfigBucket.deserialize(blob, b_v)

    # len(results) == 0
    return None


def write_exporter(user_id, account_version, store_path):
    with GetDBSession(store_path) as session:
        try:
            session.begin()
            res = session.execute(
                '''
                INSERT OR REPLACE INTO Store (bucket, bucket_version, key, value)
                VALUES (:bucket, :version, :key, :value)
                ''', {
                    'bucket': ExporterBucket.NAME,
                    'version': ExporterBucket.CURRENT_VERSION,
                    'key': ExporterBucket.key(),
                    'value': ExporterBucket.serialize(user_id, account_version, ExporterBucket.CURRENT_VERSION)
                }
            )
            session.commit()
            return res.lastrowid
        except Exception:
            session.rollback()
            raise


def read_exporter(store_path):
    with GetDBSession(store_path) as session:
        result = session.execute(
            '''
            SELECT id, value, bucket_version FROM Store
            WHERE bucket = :bucket AND key = :key
            ''', {
                'bucket': ExporterBucket.NAME,
                'key': ExporterBucket.key()
            }
        )
        r = result.fetchone()
    if r is not None:
        (_, blob, b_v) = r
        return ExporterBucket.deserialize(blob, b_v)
    return (None, None)



class SMTPConfig(object):
    def __init__(self, host, port, username, password, send_to_user, use_tls=False):
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.use_tls = use_tls
        self.send_to_user = send_to_user

    def to_json(self):
        return {
            'host': self.host,
            'port': self.port,
            'username': self.username,
            'password': self.password,
            'use_tls': self.use_tls,
            'send_to_user': self.send_to_user
        }


class IMAPConfig(object):
    def __init__(self, host, port, username, password, use_tls=False):
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.use_tls = use_tls

    def to_json(self):
        return {
            'host': self.host,
            'port': self.port,
            'username': self.username,
            'password': self.password,
            'use_tls': self.use_tls
        }


class CSConfig(object):
    def __init__(self, host, port, use_tls=False):
        self.host = host
        self.port = port
        self.use_tls = use_tls

    @property
    def http(self):
        return u'{}://{}{}'.format(
            'https' if self.use_tls else 'http',
            self.host,
            '' if self.port == 80 else ':{}'.format(self.port))

    @property
    def ws(self):
        return u'{}://{}{}'.format(
            'wss' if self.use_tls else 'ws',
            self.host,
            '' if self.port == 80 else ':{}'.format(self.port))

    def to_json(self):
        return {
            'host': self.host,
            'port': self.port,
            'use_tls': self.use_tls
        }


class Config(object):
    def __init__(self, smtp, imap, cs):
        self.smtp = smtp
        self.imap = imap
        self.cs = cs
        # more to add

    @classmethod
    def from_yaml(cls, yaml_file):
        from pvHelpers.utils import read_yaml_config

        y = read_yaml_config(yaml_file)
        return cls.from_json(y)

    @classmethod
    def from_json(cls, j):
        s, i, cs = j['smtp'], j['imap'], j['cs']
        return cls(
            SMTPConfig(s['host'], s['port'], s['username'], s['password'], s['send_to_user'], s['use_tls']),
            IMAPConfig(i['host'], i['port'], i['username'], i['password'], i['use_tls']),
            CSConfig(cs['host'], cs['port'], cs['use_tls']))

    def __repr__(self):
        return str(self.to_json())

    def to_json(self):
        return {
            'smtp': self.smtp.to_json(),
            'imap': self.imap.to_json(),
            'cs': self.cs.to_json()
        }



import click

from mail_relay.store import DEFAULT_STORE_PATH, Store

from .config import config
from .context import Context, ctx
from .info import info
from .start import start
from .status import status


@click.group()
@click.option(
    '-p', '--store-path', 'store_path',
    type=click.Path(dir_okay=False),
    default=DEFAULT_STORE_PATH, show_default=True,
    help='Path to sqlite store.')
@click.option(
    '--store-version', 'store_version',
    type=click.INT,
    default=0, show_default=True,
    help='sqlite store version.')
@click.pass_context
def driver(ctx, store_path, store_version):
    '''relay is a simple cli tool that relays preveil emails to a configurable smtp server'''

    # Initiate context object
    ctx.obj = Context(Store(store_path, store_version))


@click.command()
@ctx
def migrate(ctx):
    '''Migrate database using mail_rely/migrate.py (or optional script).'''
    # TODO dynamic migrate script
    from mail_relay.store.migrate import migrate
    migrate(ctx.store.path)


driver.add_command(info)
driver.add_command(config)
driver.add_command(migrate)
driver.add_command(start)
driver.add_command(status)



import click

from .context import ctx


@click.group(invoke_without_command=True)
@click.pass_context
def status(gctx):
    '''Check status of relay daemon'''

    if gctx.invoked_subcommand is None:
        click.echo('daemon status check')


@click.command()
@click.option('-t', '--test-send', 'test_send', flag_value=True, default=False)
@ctx
def smtp(ctx, test_send):
    '''Check smtp connection status'''
    from mail_relay.store.handlers import read_config

    config = read_config(ctx.store.path)
    if not config:
        click.echo('No config set, configure using --config-file!')

    from mail_relay.mail import test_smtp_connection, send_test_mail, create_test_mime
    click.echo('checking smtp status for \r\n {}\r\n'.format(config.smtp.to_json()))
    test_smtp_connection(config.smtp)

    if test_send:
        m = create_test_mime('saman@preveil.com', ['saman@preveil.com'])
        # TODO: check result, handle failures
        result = send_test_mail(config.smtp, m)
        print result


status.add_command(smtp)



import os


from mail_relay.mail import GetSMTPConnetion
import pytest

from ..utils import create_test_mime, crypto_create_account, rand_user_id


@pytest.mark.integration
def smtp_connection_test(test_config):
    sender_id, recip_id = rand_user_id(), rand_user_id()
    crypto_create_account(os.environ['CRYPTO_HOST'], os.environ['CRYPTO_PORT'], sender_id)
    crypto_create_account(os.environ['CRYPTO_HOST'], os.environ['CRYPTO_PORT'], recip_id)
    recipients = [recip_id]
    msg = create_test_mime(sender_id, recipients)
    with GetSMTPConnetion(test_config.smtp) as conn:
        conn.ehlo_or_helo_if_needed()
        conn.verify(sender_id)
        conn.sendmail(sender_id, recipients, msg)



from pvHelpers.crypto import PVKeyFactory
from pvHelpers.utils import b64enc, jdumps, utf8Encode

from .base_handler import SecureMessageManager
from .base_handler import SecureMessageSendingStates as States
from .base_handler import TransferTypes
from .key_exchange_manager import KeyExchangeManager


class GetMessageManager(SecureMessageManager):
    __transfer_side__ = u'connect'

    def __init__(self, relay_server_url, user_id, connect_to, transfer_type, pin_verifier, callback):
        super(GetMessageManager, self).__init__(relay_server_url, user_id, connect_to)

        self.transfer_type = transfer_type
        self.pin_verifier = pin_verifier
        self.callback = callback

        self.protocol_version = 1

        self._got_ok = False
        self._other_ephemral_public_key_serialized = None

    def _advance_state(self):
        if self.state == States.NOT_CONNECTED:
            self.state = States.WAITING_FOR_HANDSHAKE

        elif self.state == States.WAITING_FOR_HANDSHAKE:
            self.state = States.WAITING_FOR_HASH

        elif self.state == States.WAITING_FOR_HASH:
            self.state = States.WAITING_FOR_PUBLIC_KEY

        elif self.state == States.WAITING_FOR_PUBLIC_KEY:
            self.state = States.WAITING_FOR_ENCRYPTED_MESSAGE

        elif self.state == States.WAITING_FOR_ENCRYPTED_MESSAGE:
            self.state = States.DONE

        elif self.state == States.DONE:
            pass

    def get_secure_channel(self, new_device_req=None, user_req=None, useless_signature=None):
        if self.state != States.NOT_CONNECTED:
            raise ValueError('Unexpected state {}'.format(self.state))
        try:
            while True:
                if self.state == States.NOT_CONNECTED:
                    self.connect_to_relay_server()
                    self.send_handshake()
                    self._advance_state()
                elif self.state == States.WAITING_FOR_HANDSHAKE:
                    message = self.read_relay_socket()
                    common_protocol = self.resolve_common_protocol(message)
                    if common_protocol is None:
                        raise RuntimeError(u'no commmon dh protocol {}'.format(message))

                    self.init_dh(KeyExchangeManager.new(common_protocol))
                    self._advance_state()
                elif self.state == States.WAITING_FOR_HASH:
                    message = self.read_relay_socket()
                    self.other_ephemeral_public_key_hash = message['public_key_hash']

                    if self.transfer_type == TransferTypes.TRANSFER:
                        self.relay_websocket.send(jdumps({
                            'action': u'key_transfer',
                            'user_id': self.user_id,
                            'device_req': new_device_req.serialized_req,
                            'signature': b64enc(
                                self._dh_manager.ephemeral_key.signing_key.sign(
                                    utf8Encode(new_device_req.serialized_req))),
                            'message': self._dh_manager.ephemeral_key.public_user_key.serialize()
                        }))
                    else:
                        raise ValueError('Unexpected transfer type {}'.format(self.transfer_type))

                    self._advance_state()
                elif self.state == States.WAITING_FOR_PUBLIC_KEY:
                    message = self.read_relay_socket()

                    # Wait for an 'ok' (server received our pk) and their public key.
                    if message.get('message') is None:
                        self._got_ok = True
                    else:
                        self._other_ephemral_public_key_serialized = message['message']

                    # Don't advance until we got:
                    #  - server ok
                    #  - other side ephemeral key
                    if self._got_ok and self._other_ephemral_public_key_serialized is not None:
                        # make sure other side ephemeral key matches it's hash
                        key_hash = KeyExchangeManager.ephemeral_key_hash(self._other_ephemral_public_key_serialized)
                        if key_hash != self.other_ephemeral_public_key_hash:
                            raise ValueError('''Unmatching public key hashes.
                                                Initial hash {}
                                                Provided ephemeral key hash {}'''.format(
                                                    self.other_ephemeral_public_key_hash, key_hash))

                        self._dh_manager.other_ephemeral_key = PVKeyFactory.deserializePublicUserKey(
                            self._other_ephemral_public_key_serialized, is_protobuf=False)

                        status = self.pin_verifier(self._dh_manager.pin)
                        if not status:
                            raise ValueError('Pin not verified.')

                        self._advance_state()
                elif self.state == States.WAITING_FOR_ENCRYPTED_MESSAGE:
                    message = self.read_relay_socket()
                    secret_message = self._dh_manager.decrypt(message['message'])

                    if self.transfer_type == TransferTypes.TRANSFER:
                        device_req_sign = message['device_signature']
                        status = self.callback(secret_message, device_req_sign)
                    elif self.transfer_type == TransferTypes.RECOVERY:
                        device_req_sign = None
                        req_approver_sign = message['approver_signature']
                        approver_id = message['approver_id']
                        approver_key_version = message['approver_key_version']

                        status = self.callback(secret_message, req_approver_sign, approver_id, approver_key_version)
                    elif self.transfer_type == TransferTypes.EXPORT:
                        status = self.callback(secret_message)
                    else:
                        raise ValueError('Unexpected transfer type {}'.format(self.transfer_type))

                    if not status:
                        raise RuntimeError('callback failure')
                    self._advance_state()
                elif self.state == States.DONE:
                    return
        finally:
            self.stop()



import os


DEFAULT_STORE_PATH = os.path.abspath(os.path.join(os.getcwd(), 'relay.sqlite'))

class Store(object):
    def __init__(self, path, version):
        self.path = path
        self.version = version

    # TODO: internalize handlers, easening store_path provision from consumers



import os
import time

from mail_relay.secure_channel import GetMessageManager, TransferTypes
from pvHelpers.crypto import PVKeyFactory
from pvHelpers.cs_client import BACKEND_CLIENT
from pvHelpers.request import DeviceRequest
from pvHelpers.user import LocalDevice, LocalUser
from pvHelpers.utils import NOT_ASSIGNED
import pytest

from .utils import crypto_create_account, crypto_send_user_key, rand_user_id


@pytest.mark.integration
def transfer_user_account_test(test_config):
    user_id = rand_user_id()

    # create account in crypto
    crypto_create_account(os.environ['CRYPTO_HOST'], os.environ['CRYPTO_PORT'], user_id)
    send_pin_handle = crypto_send_user_key(os.environ['CRYPTO_HOST'], os.environ['CRYPTO_PORT'], user_id)

    def verifier(pin):
        # sleep needed for sockets read/write timing sequence
        time.sleep(0.1)
        send_pin_handle(pin)
        return True

    def callback(transferred_user_key, device_req_signature):
        user_key = PVKeyFactory.deserializeUserKey(transferred_user_key, is_protobuf=False)
        user = LocalUser(
            user_id, -153, user_id, u'', None, NOT_ASSIGNED(), u'', [user_key], device_req.device)

        # check user is authorized
        BACKEND_CLIENT.init(test_config.cs.http)
        BACKEND_CLIENT.getUserDevices(user)
        return True

    device_req = DeviceRequest.newForTransfer(user_id, LocalDevice.newDevice())
    manager = GetMessageManager(
        test_config.cs.ws, user_id, user_id, TransferTypes.TRANSFER, verifier, callback)

    # test blocks here till `callback` function returns
    manager.get_secure_channel(new_device_req=device_req)



from pvHelpers.crypto import PVKeyFactory
from pvHelpers.user import LocalDevice, LocalUser, OrganizationInfo
from pvHelpers.utils import b64dec, jdumps, jloads, NOT_ASSIGNED


class UserBucket:
    NAME = u'user'
    CURRENT_VERSION = 0

    @staticmethod
    def key(user_id, account_version):
        return '{};{}'.format(account_version, user_id.lower())

    @staticmethod
    def serialize(user, bucket_version):
        if bucket_version == 0:
            return UserBucket._serialize_v0(user)
        else:
            raise ValueError(u'unsupported bucket bucket_version {}'.format(bucket_version))

    @staticmethod
    def deserialize(blob, bucket_version):
        if bucket_version == 0:
            return UserBucket._deserialize_v0(blob)
        else:
            raise ValueError(u'unsupported bucket bucket_version {}'.format(bucket_version))

    @staticmethod
    def _serialize_v0(user):
        return jdumps({
            'user_id': user.user_id,
            'account_version': user.account_version,
            'display_name': user.display_name,
            'mail_cid': user.mail_cid,
            'org_info': None if user.org_info is None else {
                'id': user.org_info.org_id,
                'name': user.org_info.org_name,
                'department': user.org_info.dept_name,
                'role': user.org_info.role
            },
            'device': {
                'id': user.device.id,
                'name': user.device.name,
                'key': user.device.key.serialize(),
                'metadata': user.device.metadata,
                'status': user.device.status
            }
        })

    @staticmethod
    def _deserialize_v0(blob):
        j = jloads(blob)
        jd, jo = j['device'], j['org_info']
        d = LocalDevice(
            jd['id'], jd['name'], PVKeyFactory.userKeyFromSerializedBuffer(b64dec(jd['key'])),
            jd['metadata'], jd['status'])
        o = jo and OrganizationInfo(jo['id'], jo['name'], jo['department'], jo['role'])
        return LocalUser(
            j['user_id'], j['account_version'], j['display_name'],
            j['mail_cid'], o, NOT_ASSIGNED(), u'', [], d)



import os

from mail_relay.store.migrate import migrate
import pytest


@pytest.fixture(scope='module')
def store(tmpdir_factory):
    s = tmpdir_factory.mktemp('store')
    store_path = os.path.abspath(os.path.join(str(s), 'test.sqlite'))
    migrate(store_path)
    return store_path



import click
from mail_relay.config import Config
from pvHelpers.utils import CaseInsensitiveSet

from .context import ctx


@click.group(invoke_without_command=True)
@click.option(
    '--config-file', 'config_file',
    envvar='CONFIG_FILE',
    type=click.Path(exists=True, dir_okay=False))
@click.pass_context
def config(gctx, config_file):
    '''Show existing configiguration or re-config an existing setup.'''

    from mail_relay.store.handlers import read_config, update_config

    if gctx.invoked_subcommand is None:
        current_config = read_config(gctx.obj.store.path)
        if config_file is not None:
            if current_config is not None:
                click.echo('Current config:')
                click.echo(click.wrap_text(repr(current_config), initial_indent='>', subsequent_indent='>'))
                click.confirm('Would you like to re-config to {}?'.format(config_file))

            new_conf = Config.from_yaml(config_file)
            update_config(new_conf, gctx.obj.store.path)
        else:
            if current_config is None:
                click.echo('No config set, configure using --config-file!')
                # TODO: prompt user to input configs
            else:
                click.echo('Current config:')
                click.echo(click.wrap_text(repr(current_config), initial_indent='>', subsequent_indent='>'))


@click.command()
@click.option('--as-exporter', 'as_exporter', flag_value=True, default=False)
@click.option('-d', 'is_delete', flag_value=True, default=False)
@click.argument('user_id')
@ctx
def user(ctx, user_id, as_exporter, is_delete):
    '''Configure organization exporter account'''

    from mail_relay.store.handlers import (read_config, read_users, write_user, write_key,
                                           read_exporter, write_exporter, delete_user)
    all_users = read_users(ctx.store.path)
    if is_delete:
        delete_user(user_id, ctx.store.path)
        return

    click.echo('Configuring {} as {}'.format(user_id, 'exporter' if as_exporter else 'approver'))
    config = read_config(ctx.store.path)

    if user_id in CaseInsensitiveSet(map(lambda (v, i): i, all_users.iterkeys())):
        click.echo(
            'User {} exists locally! Please remove the user'
            ' ```relay config user -d ...``` and try again.'.format(user_id))

        return

    from mail_relay.secure_channel import GetMessageManager, TransferTypes
    from pvHelpers.crypto import PVKeyFactory
    from pvHelpers.request import DeviceRequest
    from pvHelpers.user import LocalUser, LocalDevice
    from pvHelpers.utils import NOT_ASSIGNED

    def verifier(pin):
        click.echo('Verification pin: {}'.format(pin))
        return True

    def callback(transferred_user_key, device_req_signature):
        user_key = PVKeyFactory.deserializeUserKey(transferred_user_key, is_protobuf=False)
        auth_user = LocalUser(
            user_id, -153, user_id, u'', None, NOT_ASSIGNED(), u'', [user_key], device_req.device)

        # get user information from server
        from pvHelpers.cs_client import BACKEND_CLIENT

        BACKEND_CLIENT.init(config.cs.http)
        u = BACKEND_CLIENT.fetchUsers(auth_user, [(auth_user.user_id, -1)])[auth_user.user_id]
        user = LocalUser(
                u.user_id, u.account_version, u.display_name, u.mail_cid,
                u.org_info, NOT_ASSIGNED(), u'', [user_key], device_req.device)

        write_key(user.user_id, user.account_version, user_key, ctx.store.path)
        write_user(user, ctx.store.path)
        if as_exporter is True:
            exporter_id, account_version = read_exporter(ctx.store.path)
            if exporter_id is not None:
                click.confirm(
                    '{} is current exporter, would you like to re-config setting '
                    '{} as exporter?'.format((exporter_id, account_version), (user.user_id, user.account_version)))

            write_exporter(user.user_id, user.account_version, ctx.store.path)

        return True

    device_req = DeviceRequest.newForTransfer(user_id, LocalDevice.newDevice())
    manager = GetMessageManager(
        config.cs.ws, user_id, user_id, TransferTypes.TRANSFER, verifier, callback)

    # test blocks here till `callback` function returns
    manager.get_secure_channel(new_device_req=device_req)


@click.command()
@click.argument('user_id')
@click.pass_context
def exporter(ctx, user_id):
    '''Configure organization export group expoter account'''

    ctx.invoke(user, user_id=user_id, as_exporter=True)


@click.command()
@click.argument('user_id')
@click.pass_context
def approver(ctx, user_id):
    '''Configure organization export group approver accounts'''

    ctx.invoke(user, user_id=user_id, as_exporter=False)


config.add_command(exporter)
config.add_command(approver)
config.add_command(user)



# -*- coding: utf-8 -*-
import os

from setuptools import find_packages, setup


def get_readme_content():
    with open(os.path.join(os.path.dirname(__file__), 'docs', 'development.md'), 'r') as f:
        return f.read()


def get_license_content():
    with open(os.path.join(os.path.dirname(__file__), 'LICENSE'), 'r') as f:
        return f.read()


setup(
    name='mail-relay',
    version='0.1.6',
    description='Just a tool to relay incoming emails over to other mail servers',
    long_description=get_readme_content(),
    long_description_content_type='text/markdown',
    author='Saman Tehrani',
    author_email='samanrtehrani@gmail.com',
    url='https://github.com/samantehrani/mail-relay',
    license='GNU General Public License v3.0',
    package_dir={'': 'src'},
    entry_points={
      'console_scripts': [
        'relay = mail_relay.cli:driver'
      ]
    },
    packages=find_packages(where='src', exclude=('tests')),
    python_requires='>=2.7'
)



from pvHelpers.store import GetDBSession


def migrate(wd):
    with GetDBSession(wd) as s:
        with s.begin():
            s.execute('DROP TABLE IF EXISTS Store')
            s.execute(
                '''
                CREATE TABLE Store
                    (id INTEGER PRIMARY KEY AUTOINCREMENT,
                    bucket TEXT NOT NULL COLLATE NOCASE,
                    bucket_version INTEGER NOT NULL,
                    key TEXT NOT NULL COLLATE NOCASE,
                    value BLOB NOT NULL,
                    UNIQUE (bucket, key))
                '''
            )



from pvHelpers.crypto import PVKeyFactory
from pvHelpers.utils import b64dec


class KeyBucket:
    NAME = u'key'
    CURRENT_VERSION = 0

    @staticmethod
    def key(user_id, account_version, key_version):
        return '{};{};{}'.format(key_version, account_version, user_id.lower())

    @staticmethod
    def serialize(user_key, bucket_version):
        if bucket_version == 0:
            return KeyBucket._serialize_v0(user_key)
        else:
            raise ValueError(u'unsupported bucket bucket_version {}'.format(bucket_version))

    @staticmethod
    def deserialize(blob, bucket_version):
        if bucket_version == 0:
            return KeyBucket._deserialize_v0(blob)
        else:
            raise ValueError(u'unsupported bucket bucket_version {}'.format(bucket_version))

    @staticmethod
    def _serialize_v0(user_key):
        return user_key.serialize()

    @staticmethod
    def _deserialize_v0(blob):
        return PVKeyFactory.userKeyFromSerializedBuffer(b64dec(blob))



import click
from mail_relay.store.handlers import read_config, read_users
from pvHelpers.cs_client import BACKEND_CLIENT
from pvHelpers.utils import CaseInsensitiveSet

from .context import ctx


@click.command()
@click.option('--org-info', 'show_org_info', flag_value=True, default=False)
@click.option('--export-group', 'show_export_group', flag_value=True, default=False)
@click.option('--members', 'show_members', flag_value=True, default=False)
@ctx
def info(ctx, show_org_info, show_export_group, show_members):
    '''Show information about organization, export groups and members.'''

    config = read_config(ctx.store.path)
    BACKEND_CLIENT.init(config.cs.http)

    users = [u for ((v, i), u) in read_users(ctx.store.path).iteritems() if v == -1]
    if len(users) == 0:
        click.echo('No users are configured! Try configuring exporter '
                   '```relay config exporter ...``` and try again.')
        return

    org_ids = CaseInsensitiveSet(
        filter(lambda i: i is not None, map(lambda u: u.org_info and u.org_info.org_id, users)))

    assert len(org_ids) == 1

    user, org_id = users[0], list(org_ids)[0]

    org_info = BACKEND_CLIENT.getOrgInfo(user, org_id)
    if (show_org_info or show_export_group or show_members) is False:  # show all if not specified
        click.echo(org_info)
    else:
        if show_org_info:  # get org info
            # latest user versions
            o = dict(org_info)
            del o['users']
            click.echo(o)

        # get export group info
        # get approvers/exporter info
        if show_export_group:
            if org_info['roled_approval_groups'].get('export_approval_group') is None:
                click.echo('No export group configured! Make sure to create and '
                           'set an export group for your organization and try again.')
                return
            group_id = org_info['roled_approval_groups']['export_approval_group']['group_id']
            version = org_info['roled_approval_groups']['export_approval_group']['version']
            export_group = BACKEND_CLIENT.getOrgAPGInfo(user, org_id, group_id, version)
            click.echo(export_group)

        # get members info
        if show_members:
            click.echo(org_info['users'])



import time

import click
from mail_relay.store.handlers import read_users

from .context import ctx


@click.command()
@ctx
def start(ctx):
    '''Start relay daemon.'''

    while True:
        try:
            print read_users(ctx.store.path)
        except Exception as e:
            print e
        finally:
            time.sleep(1)
